import 'dart:convert';

import 'package:faiadashu/resource_provider/resource_provider.dart';
import 'package:fhir/r4.dart';

/// Provide commonly used [ValueSet]s in hard-coded form.
///
/// As per the FHIR standard, this has to include the following Value Sets:
/// * 'http://hl7.org/fhir/ValueSet/yesnodontknow'
///
/// see: http://build.fhir.org/questionnaire.html#valuesets
class FhirValueSetProvider extends FhirResourceProvider {
  @override
  ValueSet? getResource(String uri) {
    switch (uri) {
      case 'http://hl7.org/fhir/ValueSet/yesnodontknow':
        return ValueSet.fromJson(
          jsonDecode(yesNoDontKnowValueSetRaw) as Map<String, dynamic>,
        );
      default:
        return null;
    }
  }

  static String yesNoDontKnowValueSetRaw = r'''
{
  "resourceType": "ValueSet",
  "id": "yesnodontknow",
  "text": {
    "status": "generated",
    "div": "<div xmlns=\"http://www.w3.org/1999/xhtml\">\n      <h3>Value Set Contents</h3>\n      <p>This value set contains 3 concepts</p>\n      <table class=\"codes\">\n        <tr>\n          <td style=\"white-space:nowrap\">\n            <b>Code</b>\n          </td>\n          <td>\n            <b>System</b>\n          </td>\n          <td>\n            <b>Display</b>\n          </td>\n          <td>\n            <b>Definition</b>\n          </td>\n        </tr>\n        <tr>\n          <td style=\"white-space:nowrap\">\n            <a name=\"http---terminology.hl7.org-CodeSystem-v2-0136-Y\"> </a>\n            <a href=\"v2/0136/index.html#v2-0136-Y\">Y</a>\n          </td>\n          <td>http://terminology.hl7.org/CodeSystem/v2-0136</td>\n          <td>Yes</td>\n          <td/>\n        </tr>\n        <tr>\n          <td style=\"white-space:nowrap\">\n            <a name=\"http---terminology.hl7.org-CodeSystem-v2-0136-N\"> </a>\n            <a href=\"v2/0136/index.html#v2-0136-N\">N</a>\n          </td>\n          <td>http://terminology.hl7.org/CodeSystem/v2-0136</td>\n          <td>No</td>\n          <td/>\n        </tr>\n        <tr>\n          <td style=\"white-space:nowrap\">\n            <a name=\"http---terminology.hl7.org-CodeSystem-data-absent-reason-asked-unknown\"> </a>\n            <a href=\"codesystem-data-absent-reason.html#data-absent-reason-asked-unknown\">asked-unknown</a>\n          </td>\n          <td>http://terminology.hl7.org/CodeSystem/data-absent-reason</td>\n          <td>Don't know</td>\n          <td>The source was asked but does not know the value.</td>\n        </tr>\n      </table>\n    </div>"
  },
  "url": "http://hl7.org/fhir/ValueSet/yesnodontknow",
  "version": "4.0.1",
  "name": "Yes/No/Don't Know",
  "status": "draft",
  "description": "For Capturing simple yes-no-don't know answers",
  "compose": {
    "include": [
      {
        "valueSet": [
          "http://terminology.hl7.org/ValueSet/v2-0136"
        ]
      },
      {
        "system": "http://terminology.hl7.org/CodeSystem/data-absent-reason",
        "concept": [
          {
            "code": "asked-unknown",
            "display": "Don't know"
          }
        ]
      }
    ]
  },
  "expansion": {
    "identifier": "urn:uuid:bf99fe50-2c2b-41ad-bd63-bee6919810b4",
    "timestamp": "2015-07-14T10:00:00Z",
    "contains": [
      {
        "system": "http://terminology.hl7.org/CodeSystem/v2-0136",
        "code": "Y",
        "display": "Yes"
      },
      {
        "system": "http://terminology.hl7.org/CodeSystem/v2-0136",
        "code": "N",
        "display": "No"
      },
      {
        "system": "http://terminology.hl7.org/CodeSystem/data-absent-reason",
        "code": "asked-unknown",
        "display": "Don't know"
      }
    ]
  }
}
''';

  @override
  Future<void> init() async {
    // Do nothing
    return;
  }

  @override
  FhirResourceProvider? providerFor(String uri) {
    return (uri == 'http://hl7.org/fhir/ValueSet/yesnodontknow') ? this : null;
  }
}
